#test output of get_player_match_data. See if all the headers are all right. See if the data received is not empty.
from xgpy.constants import TEST_PLAYER_ID, TEST_FILTER, TEST_SEASON, TEST_LEAGUE
from xgpy.understat import UnderstatPlayer


class TestUnderstatPlayer():

    def test_get_player_match_data(self):

        tp = UnderstatPlayer(TEST_PLAYER_ID)
        data = tp.get_player_match_data(**TEST_FILTER)

        assert isinstance(data, list) and len(data) > 0

    def test_get_player_shot_data(self):

        tp = UnderstatPlayer(TEST_PLAYER_ID)
        data = tp.get_player_shot_data(**TEST_FILTER)

        assert isinstance(data, list) and len(data) > 0

    def test_get_player_grouped_data_by_type(self):

        tp = UnderstatPlayer(TEST_PLAYER_ID)
        data = tp.get_player_grouped_data_by_type(**TEST_FILTER)

        assert isinstance(data, list) and len(data) > 0

    def test_get_player_min_max_data(self):

        tp = UnderstatPlayer(TEST_PLAYER_ID)
        data = tp.get_player_min_max_data()

        assert isinstance(data, dict) and len(data) > 0

    def test_get_player_list_by_league(self):

        data = UnderstatPlayer.get_player_list_by_league(TEST_LEAGUE, TEST_SEASON)

        assert isinstance(data, dict) and len(data) > 0



from xgpy.constants import TEST_TEAM, TEST_LEAGUE, TEST_SEASON
from xgpy.understat import UnderstatTeam

class TestUnderstatTeam():

    def test_get_team_league_history(self):

        # test that the data received is of type dictionary and not empty

        tp = UnderstatTeam(TEST_TEAM, TEST_LEAGUE)
        data = tp.get_team_league_history(TEST_SEASON)

        assert isinstance(data, dict) and len(data) > 0

    def test_get_team_player_summary(self):

        # test that the data received is of type dictionary and not empty

        tp = UnderstatTeam(TEST_TEAM, TEST_LEAGUE)
        data = tp.get_team_player_summary(TEST_SEASON)

        assert isinstance(data, list) and len(data) > 0

    def test_get_team_player_summary(self):

        # test that the data received is of type dictionary and not empty

        tp = UnderstatTeam(TEST_TEAM, TEST_LEAGUE)
        data = tp.get_team_fixtures(TEST_SEASON)

        assert isinstance(data, list) and len(data) > 0

    def test_get_team_player_summary(self):

        # test that the data received is of type dictionary and not empty

        tp = UnderstatTeam(TEST_TEAM, TEST_LEAGUE)
        data = tp.get_team_grouped_date_by_type(TEST_SEASON)

        assert isinstance(data, dict) and len(data) > 0



from xgpy.constants_fbref import *
from xgpy.Utility import Utility
from bs4 import BeautifulSoup
import json
import pandas as pd

class fbrefPlayer():

    """
    A class dedicated to fetch information about a player
    from fbref.

    """

    def __init__(self, player_id):

        # store the player id
        self.id = player_id

    def get_player_aggregate_stats(self, type, competition='dom_lg'):

        """
        get aggregate career statistics of player.

        :param type: this is the type of statistics that are to be fetched. possible values:
            - standard
            - shooting
            - passing
            - passing_types
            - goal_shot_creation
            - defensive_actions
            - possession
            - playing_time
            - miscellaneous
        :type type: str
        :param competition: the competition for which stats are to be fetched. possible values:
            - all_competitions
            - domestic_league
            - domestic_cup
            - internationl_cup
            - national_team
        """

        # TODO: there seems to be something wrong with /all_comps/ endpoint
        # TODO: in all_comps endpoint, not able to fetch other than standard stats. commented out.
        # TODO: need to add error appropriate error handling when accessing dictionaries.
        # TODO: format the output data with column names and remove unncessary columns (matches)

        main_url = Utility.generate_request_url(PLAYER_URL, *(self.id, FBREF_COMPETITION_TO_URL_MAP[competition]))
        r = Utility.generate_request_object(main_url)
        print(main_url)

        if type not in FBREF_STATS_TO_CLASS_MAP.keys():
            raise ValueError('{} is not a valid stat type. Please refer to documentation for stat types.'.format(type))

        if competition not in FBREF_COMPETITION_TO_URL_MAP.keys():
            raise ValueError('{} is not a valid competition. Please refer to documentation for supported competitions.'.format(type))

        data = Utility.find_and_get_soup_table(r, FBREF_STATS_TO_CLASS_MAP[type], FBREF_COMPETITION_TO_URL_MAP[competition])
        df = pd.DataFrame(data)
        df.columns = df.iloc[0]
        df = df.drop(df.index[0])

        return df



import requests
import re
from xgpy.constants import *
import json
import pandas as pd
from bs4 import BeautifulSoup


class Utility():

    """
    A class comprising of all the helper functions.

    Methods
    -------
    generate_request_url(base_url, param)
        construct the URL to be queried.

    generate_request_object(url)
        generate a request object for the url specified.

    string_escape(s, encoding='utf-8')
        convert the string from given encoding (default utf-8) to human readable text.

    find_match(request, match_string)
        find a match in the request based on the match string provided.

    build_and_match(base_url, id, **filter)
        build and connect to the url and find desired data

    """


    @staticmethod
    def generate_request_url(base_url: str, *param: str):
        """
        Parameters
        ----------
        base_url: str
            the core understat url for the related endpoint. see constants.py for example.
        param: str
            value to format the string with (team or player id).

        Returns
        --------
        str
            a url string to fetch the data from.
        """
        return base_url.format(*param)

    @staticmethod
    def generate_request_object(url: str):
        """
        Parameters
        ---------
        url: str
            the url to which a connection is to be made.

        Raises
        ------
        Exception
            A generic exception when a connection cannot be established.

        """

        try:
            return requests.get(url)
        except Exception:
            # TODO: Add a more detailed explanation
            print("An error occured")

    @staticmethod
    def string_escape(s, encoding='utf-8'):
        """
        Parameters
        ----------
        s: str
            the string to be decoded.
        encoding: str
            the encoding with which the string is to be converted.

        Returns
        -------
        str
            a human readable string

        """

        return (s.encode('latin1')         # To bytes, required by 'unicode-escape'
                # Perform the actual octal-escaping decode
                .decode('unicode-escape')
                .encode('latin1')         # 1:1 mapping back to bytes
                .decode(encoding))        # Decode original encoding

    @staticmethod
    def find_match(request, match_string: str):
        """
        Parameters
        ----------
        request: object
            a request object that sucessfully connected to the desired URL.
        match_string: str
            the string used to find a match. Depends on the type of data being queried for. see constants.

        Returns
        -------
        Match
            returns a regex match object
        """

        match_pattern = re.compile(DATA_PATTERN.format(match_string))
        match = re.search(match_pattern, request.text)

        return match

    @staticmethod
    def build_and_match(base_url: str, search_keyword: str, *url_params: list):
        """
        Parameters
        ----------
        base_url: str
            the base_url from where information is to be queried.
        id: int
            the understat played_id.
        search_keyword: str
            the keyword search for the data to be fetched.

        Returns
        -------
        str:
            returns JSON data in a human readable format
        """

        base_url = Utility.generate_request_url(base_url, *url_params)
        r = Utility.generate_request_object(base_url)
        match = Utility.find_match(r, search_keyword)
        string_data = Utility.string_escape(match.group(1))

        return string_data

    @staticmethod
    def filter_data(json_data, **filter):
        """
        Parameters
        ----------
        base_url: str
            the base_url from where information is to be queried.
        id: int
            the understat played_id.
        search_keyword: str
            the keyword search for the data to be fetched.

        Returns
        -------
        str:
            returns JSON data in a human readable format
        """

        # TODO: need to convert filtering to a function
        season = filter.get('season', '')

        # if a filter exists, do something
        if season:
            json_data = [x for x in json_data if x['season'] == season]

        return json_data

    @staticmethod
    def find_and_get_soup_table(r, mapped_type, mapped_competition):
        """
        get the data based on a the stat and the competition

        :param mapped_type: the stat type converted into fbref format
        :type mapped_type: str
        :param mapped_competition: the competition converted into fbref format
        :type mapped_competition: str

        :return: specified stat data for competition of player
        :rtype: dict

        """


        soup = BeautifulSoup(r.text, 'html.parser')                     # Create bs4 html parser
        attribute_key = '_'.join([mapped_type, mapped_competition])     # create the attribute which will be used to find our data

        table = soup.find('table', attrs={
            'id': attribute_key                                         # find the data
        })

        # in case data not present, raise error to double check arguments
        if not table:
            raise ValueError('Please enter a valid stat type and competition or check if stat exists at {}'.format(r.url))


        # find the header information

        data = []
        table_head = table.find('thead')

        row = table_head.find_all('tr')[-1]
        cols = row.find_all('th')
        cols = [ele.text.strip() for ele in cols]
        data.append([ele for ele in cols])


        # get body of table
        # iterate through the rows of the table.
        # append everything in a line to an array.
        # append the line to another array.
        table_body = table.find('tbody')

        rows = table_body.find_all('tr')
        for row in rows:
            cols = row.find_all([
                    'th',
                    'td'
            ])
            cols = [ele.text.strip() for ele in cols]
            data.append([ele for ele in cols])

        return data



from xgpy.constants import *
from xgpy.Utility import Utility
import json
import pandas as pd


class UnderstatPlayer():

    """
    A class that represents an understat football player logically.

    Attributes
    ----------
    id : int
        the id of the player instance in understat.

    """

    def __init__(self, player_id):

        """
        intializer for the UnderstatPlayer class.

        :param id: the id of the player who's data is needed.
        :type id: int
        """

        self.id = player_id

    def get_player_match_data(self, **filter):

        """
        get xG data for the player.

        :param \*\*filter: a dictionary of possible filters.
        :type \*\*filter: dict
        :return: match data of the player
        :rtype: list
        """

        # First generate the URL from which data will be fetched
        # Next, generate a request object based on the URL
        # Finally, using regular expressions, finding the json
        # data using the keyword for the query
        string_data = Utility.build_and_match(PLAYER_URL, PLAYER_MATCHES_DATA, self.id)

        # TODO: What if empty?
        json_data = json.loads(string_data)

        # TODO: filter data
        json_data = Utility.filter_data(json_data, **filter)

        return json_data

    def get_player_grouped_data_by_type(self, type = 'season', **filter):

        """
        get the per season data for the player.

        :param type: the type of data to be fetched.
        :type type: str
        :param \*\*filter: a dictionary of possible filters.
        :type \*\*filter: dict
        :return: match data of the player
        :rtype: list
        """

        # First generate the URL from which data will be fetched
        # Next, generate a request object based on the URL
        # Finally, using regular expressions, finding the json
        # data using the keyword for the query
        string_data = Utility.build_and_match(PLAYER_URL, PLAYER_GROUPED_DATA, self.id)

        # TODO: What if empty?
        json_data = json.loads(string_data)

        # TODO: Add error handling here
        json_data = json_data[type]

        # TODO: filter data
        json_data = Utility.filter_data(json_data, **filter)

        return json_data

    def get_player_shot_data(self, **filter):

        """
        get the shot data for the player.

        :param \*\*filter: a dictionary of possible filters.
        :type \*\*filter: dict
        :return: shot data of the player
        :rtype: list
        """

        # First generate the URL from which data will be fetched
        # Next, generate a request object based on the URL
        # Finally, using regular expressions, finding the json
        # data using the keyword for the query
        string_data = Utility.build_and_match(PLAYER_URL, PLAYER_SHOT_DATA, self.id)

        # TODO: What if empty?
        json_data = json.loads(string_data)

        # TODO: filter data
        json_data = Utility.filter_data(json_data, **filter)

        return json_data

    def get_player_min_max_data(self, **filter):

        """
        get the min, max and average data for the player.

        :param \*\*filter: a dictionary of possible filters.
        :type \*\*filter: dict
        :return: min max data of the player
        :rtype: list
        """

        # First generate the URL from which data will be fetched
        # Next, generate a request object based on the URL
        # Finally, using regular expressions, finding the json
        # data using the keyword for the query
        string_data = Utility.build_and_match(PLAYER_URL, PLAYER_MIN_MAX_DATA, self.id)

        # TODO: What if empty?
        json_data = json.loads(string_data)

        # TODO: filter data
        json_data = Utility.filter_data(json_data, **filter)

        return json_data

    def get_player_positions(self, **filter):

        """
        get the possble player locations

        :return: list of positions of the player
        :rtype: list
        """

        # TODO: currently, all possible positions (even GK for an outfield player) are being listed
        # would be better if we return just the positions played so far

        # First generate the URL from which data will be fetched
        # Next, generate a request object based on the URL
        # Finally, using regular expressions, finding the json
        # data using the keyword for the query
        string_data = Utility.build_and_match(PLAYER_URL, PLAYER_POSITIONS, self.id)

        # TODO: What if empty?
        json_data = json.loads(string_data)

        json_data = Utility.filter_data(json_data, **filter)
        return json_data

    @staticmethod
    def get_player_list_by_league(league, season):

        """
        get all possible players in a league.

        :param league: the league for which the data is to be fetched. multiple values can be passed seperated by a ',' .
            possible values:
                - EPL
                - La_Liga
                - Bundesliga
                - Serie_A
                - Ligue_1
                - RFPL
        :type league: str
        :param season: season for which data is needed. For 2020/21 season, enter 2020.
        :type season int:
        :return: players in the league listed.
        :rtype: dict
        """

        # TODO: there seems to be an issue with certain player names
        # special characters mess up

        # get all parameters seperately. clean any dangling whitespaces.
        all_leagues = [x.strip() for x in league.split(',')]

        # get data for every league. get only select parameters.
        # append league name. concat all leagues and return as dict
        data = []
        for league in all_leagues:
            json_data = Utility.build_and_match(LEAGUE_URL, PLAYER_LIST_DATA, *(league, season))
            clean_data = pd.DataFrame(eval(json_data)).loc[:, PLAYER_DATA_COLUMNS]
            clean_data['league'] = league

            data += [clean_data]

        return pd.concat(data).to_dict()


class UnderstatTeam():
    """
    A class to get data based on a team.

    """
    def __init__(self, team_name, league):
        self.team_name = team_name
        self.league = league

    def get_team_league_history(self, season, **filter):

        """
        get the league history for a given team

        :param season: season for which history is needed. For 2020/2021 season, input 2020.
        :type season: str
        :return: dictionary containing the data for matches played in given season
        :rtype: dict
        """

        string_data = Utility.build_and_match(LEAGUE_URL, TEAMS_STANDINGS_DATA, *(self.league, season))
        json_data = json.loads(string_data)

        json_data = Utility.filter_data(json_data, **filter)

        # get data only for the team instance
        for key in json_data.keys():
            if json_data[key]["title"] == ' '.join(self.team_name.split('_')):
                return json_data[key]

    def get_team_player_summary(self, season, **filter):

        """
        get the summary statistics for each player of the team for a given season.

        :param season: season for which history is needed. For 2020/2021 season, input 2020.
        :type season: str
        :return: list containing summary statistics of players
        :rtype: list
        """

        string_data = Utility.build_and_match(TEAM_URL, PLAYER_LIST_DATA, *(self.team_name, season))
        json_data = json.loads(string_data)

        json_data = Utility.filter_data(json_data, **filter)
        return json_data

    def get_team_fixtures(self, season, **filter):

        """
        get the team fixtures for given season. for the fixtures that have been played, basic stats
        are included.

        :param season: season for which fixtures are needed. For 2020/2021 season, input 2020.
        :type season: str
        :return: list containing fixtures
        :rtype: list

        """

        string_data = Utility.build_and_match(TEAM_URL, TEAMS_FIXTURES_DATA, *(self.team_name, season))
        json_data = json.loads(string_data)

        json_data = Utility.filter_data(json_data, **filter)
        return json_data

    def get_team_grouped_date_by_type(self, season, type='situation', **filter):

        """
        get the per season data for the team.

        :param season: season for which fixtures are needed. For 2020/2021 season, input 2020.
        :type season: str
        :param type: the type of data to be fetched. This can be one of the following:
            - situation
            - formation
            - gameState
            - timing
            - shotZone
            - attackSpeed
        :type type: str
        :return: dictionary containing fixtures
        :rtype: dict
        """

        string_data = Utility.build_and_match(TEAM_URL, TEAMS_GROUPED_DATA, *(self.team_name, season))
        json_data = json.loads(string_data)

        json_data = Utility.filter_data(json_data, **filter)
        return json_data

class UnderstatMatch():

    """
    A class to model a match between two teams and obatin information.
    """

    def __init__(self, match_id):
        self.id = match_id

    def get_match_shot_data(self):

        """
        get the shot data for the match.

        :return: list containing shot data for the match.
        :rtype: list
        """

        string_data = Utility.build_and_match(MATCH_URL, MATCH_SHOT_DATA, self.id)
        json_data = json.loads(string_data)

        return json_data

    def get_match_stats(self):

        """
        get the general match statistics for both teams.

        :return: a dictionary containing match statistics.
        :rtype: dict

        """

        string_data = Utility.build_and_match(MATCH_URL, MATCH_INFO_DATA, self.id)
        json_data = json.loads(string_data)

        return json_data

    def get_match_roster(self):

        """
        get the match roster for both teams.

        :return: a dictionary containing match roster for both teams.
        :rtype: dict

        """

        string_data = Utility.build_and_match(MATCH_URL, MATCH_ROSTER_DATA, self.id)
        json_data = json.loads(string_data)

        return json_data



from xgpy.constants import TEST_MATCH, TEST_FILTER, TEST_SEASON, TEST_LEAGUE
from xgpy.understat import UnderstatMatch

class TestUnderstatMatch():

    def test_get_match_shot_data(self):

        tp = UnderstatMatch(TEST_MATCH)
        data = tp.get_match_shot_data()

        assert isinstance(data, dict) and len(data) > 0

    def test_get_match_stats(self):

        tp = UnderstatMatch(TEST_MATCH)
        data = tp.get_match_stats()

        assert isinstance(data, dict) and len(data) > 0

    def test_get_match_roster(self):

        tp = UnderstatMatch(TEST_MATCH)
        data = tp.get_match_roster()

        assert isinstance(data, dict) and len(data) > 0



# Main URL's to fetch data
PLAYER_URL = 'https://fbref.com/en/players/{}/{}/'

# statistics class mapping
FBREF_STATS_TO_CLASS_MAP = {
    'standard': 'stats_standard',
    'shooting': 'stats_shooting',
    'passing':  'stats_passing',
    'passing_types': 'stats_passing_types',
    'goal_shot_creation': 'stats_gca',
    'defensive_actions': 'stats_defense',
    'possession': 'stats_possession',
    'playing_time': 'stats_playing_time',
    'miscellaneous': 'stats_misc'
}

# This is used to generate the competition
# part of the class id from which data is to be identified

FBREF_COMPETITION_TO_KEY_MAP = {
    'all_competitions': 'ks_collapsed',
    'domestic_league': 'dom_lg',
    'domestic_cup': 'dom_cup',
    'internationl_cup': 'intl_cup',
    'national_team': 'nat_tm'
}

# This is used to convert the competition
# argument into a parameter for the URL
FBREF_COMPETITION_TO_URL_MAP = {
    'all_competitions': 'all_comps',
    'domestic_league': 'dom_lg',
    'domestic_cup': 'dom_cup',
    'internationl_cup': 'intl_cup',
    'national_team': 'nat_tm'
}



from xgpy.Utility import Utility
from xgpy.constants import PLAYER_URL, TEST_PLAYER_ID, TEST_PLAYER_URL, TEST_STR

class TestUtility():

    def test_generate_url_player(self):

        # assert that the url generated is as intended
        assert Utility.generate_request_url(PLAYER_URL, TEST_PLAYER_ID) == TEST_PLAYER_URL

    def test_generate_request_object(self):
        url = Utility.generate_request_url(PLAYER_URL, TEST_PLAYER_ID )

        # assert that the request went through well
        assert Utility.generate_request_object(url).status_code == 200

    def test_string_escape(self):

        # assert that the string conversion actually works
        assert Utility.string_escape(TEST_STR) == "{"



# MAIN URL's
PLAYER_URL = "https://understat.com/player/{}"
LEAGUE_URL = "https://understat.com/league/{}/{}/"
TEAM_URL = "https://understat.com/team/{}/{}"
MATCH_URL = "https://understat.com/match/{}"

# Pattern of the JSON variables
DATA_PATTERN = r"{}\s+=\s+JSON.parse\(\'(.*)\'\)"

# JSON_CONSTANTS: The variable names which prefix the data
PLAYER_MATCHES_DATA = "matchesData"
PLAYER_GROUPED_DATA = "groupsData"
PLAYER_SHOT_DATA = "shotsData"
PLAYER_MIN_MAX_DATA = "minMaxPlayerStats"
PLAYER_POSITIONS = "positionsList"
PLAYER_LIST_DATA = "playersData"
TEAMS_STANDINGS_DATA = "teamsData"
TEAMS_FIXTURES_DATA = "datesData"
TEAMS_GROUPED_DATA = "statisticsData"
MATCH_SHOT_DATA = "shotsData"
MATCH_INFO_DATA = "match_info"
MATCH_ROSTER_DATA = "rostersData"

# Filter constants
ARG_SEASON = "season"
ARG_START_DATE = 'start_date'
ARG_END_DATE = 'end_date'
ARG_POSITION = 'position'
PLAYER_DATA_COLUMNS = [  # for get_player_list_by_league
        'id',
        'player_name',
        'team_title'
    ]

# Variables for testing
TEST_FILTER = {
    "season": '2020',
}
TEST_PLAYER_ID = 1228
TEST_PLAYER_URL = "https://understat.com/player/{}".format(TEST_PLAYER_ID)
TEST_SEASON = 2020
TEST_LEAGUE = 'EPL'
TEST_TEAM = 'Manchester_United'
TEST_STR = '\x7B'
TEST_MATCH = 14596



from understat import UnderstatPlayer
from Utility import Utility
from constants import *
import json

a = UnderstatPlayer(1228)

filter = {
    'season': '2020'
}
t = a.get_player_match_data(**filter)
print(t)
# print(Utility.generate_request_object(Utility.generate_request_url(PLAYER_URL, TEST_PLAYER_ID)).status_code)
# Utility.filter_data({'ranjan':2020})
